/* Free for absolutely any use whatsoever. 
**   Module: midihelper 
**   Author: Thomas Moore (garavondik at yahoo.com)
**   Message: My original intent for this was to be a generic file helper
**	    but Josh smacked me down on the forums. So here is my midi 
**	    helper. I suck at C so don't hurt me if I blow up
**	    your iPod.
*/

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fnmatch.h>
#include <signal.h>
#include <linux/soundcard.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include "pz.h"

#ifndef MAXPATHLEN
#define MAXPATHLEN 512
#endif

static PzModule *midi_module;
static TWindow *open_midi(const char *filename);
static TWindow *ret;
static PzWidget *wid;

static pid_t kidpid;
static char *text = "iPod TiMidity version 0.1\n \
(C) 1995 Tuukka Toivonen\n \
(C) 2006 port by mrscolex\n \
Hit menu to exit\n";

static short int internalvolume=100;
static int return_value;

static pz_dsp_st *snd_dev;

static void wait_on_sigchld(int signal)
{ 
	wait(NULL);
}

void setup_sigchld_handler()
{
        static char sigchld_handler;
        struct sigaction act;

        if (sigchld_handler) return;

        act.sa_handler = wait_on_sigchld;
        act.sa_flags = 0;
        sigaction(SIGCHLD, &act, NULL);
        sigchld_handler = 1;
}

void draw_midihelper (PzWidget *wid, ttk_surface srf) 
{
	char *p = strdup(text);
	char tmp[20];

	int y = wid->y + 5;

	while (*p) {
	        char svch;
	        if (strchr (p, '\n')) {
	            svch = *strchr (p, '\n');
	            *strchr (p, '\n') = 0;
	        } else {
	            svch = 0;
	        }
		ttk_text (srf, ttk_textfont, wid->x + 5, y, ttk_makecol (BLACK), p);
	        p += strlen (p);
	        y += ttk_text_height (ttk_textfont) + 1;
	        *p = svch;
	        if (*p) p++;
	}
	y += 3;
	/* Draw a line. */
	ttk_line (srf, wid->x + 5, y, wid->x + wid->w - 5, y, ttk_makecol (DKGREY));

	sprintf(tmp, "Volume: %d%%", internalvolume);
	ttk_text (srf, ttk_menufont, wid->x + (wid->w - ttk_text_width (ttk_menufont, tmp)) / 2,
              wid->y + wid->h - ttk_text_height (ttk_menufont) - 5, ttk_makecol (BLACK), tmp);
}


int event_midihelper (PzEvent *ev) 
{
	/* This is our button handler. */
    
    switch (ev->type) {
	    case PZ_EVENT_SCROLL:
		if (ev->arg < 0) {
			if ((internalvolume - 10) > -1)
			{
				internalvolume=internalvolume-10;
				kill(kidpid, SIGUSR1);
			}
			break;
		}
		else {
			if ((internalvolume + 10) < 101)
			{
				internalvolume=internalvolume+10;
				kill(kidpid, SIGUSR2);
			}
			break;
		}
	    case PZ_EVENT_BUTTON_DOWN:
		switch (ev->arg) {
			/* If they hit menu, kill the midi and exit */
			case PZ_BUTTON_MENU:
				pz_close_window (ev->wid->win);
				kill(kidpid, SIGQUIT);
				break;
			default:
				break;
		}
	    case PZ_EVENT_DESTROY:
		break;
    }
    return 0;
}

static int is_midi(const char *filename)
{

/* Piece of cake */

	if (fnmatch("*.mid", filename, FNM_PATHNAME) == 0 || 
            fnmatch("*.MID", filename, FNM_PATHNAME) == 0)
		return 1;
	else
		return 0;
}

static int midihelper_loop(struct TWidget *this)
{
	int vol, mixed;

	if (waitpid(kidpid, &return_value, WNOHANG) != 0)
	{
		pz_close_window (this->win);
		kill(kidpid, SIGQUIT);
	}

	this->dirty = 1;
	return 0;
}

static TWindow *open_midi(const char *filename)
{
    struct stat buf;

/* Make sure that the prerequisite files exist */

   if (lstat("/hp/midi_instruments", &buf) == -1)
   {
      pz_error("Midi Instruments directory does not exist!");
      return TTK_MENU_DONOTHING;
   }

   if (lstat("/usr/share/timidity/timidity.cfg", &buf) == -1)
   {
      pz_error("Midi configuration file does not exist!");
      return TTK_MENU_DONOTHING;
   }

   if (lstat("/usr/bin/timidity", &buf) == -1)
   {
      pz_error("Midi executable does not exist!");
      return TTK_MENU_DONOTHING;
   }

/* Setup signal handling for children pids */
    setup_sigchld_handler();


/* This is where I suck big time. *shrug*.
** I don't want to hear about how sloppy this all is ;) 
** Arguments for TiMiDiTy, set for low quality mostly
*/

/* Josh's sexy code, blame him if it doesn't work */

	switch ((kidpid = vfork())) 
	{ 
	 case 0: /* Because directory traversal was dumb
                 ** Depending on where the File Browser was
                 ** and module_get_datapath only returns a relative path
                 ** from podzilla's root, I decided it was better
		 ** to distribute a binary for midi playing.
		 ** Thus the hardcoding of paths below.
                 */

/* My iBook only plays 8 bit sound through /dev/dsp 
** so yes this is a bit heavy handed to force 8-bit through timidity
** but it's really obnoxious to hear static... and its only for non
** Desktop pz2 builds/users.
*/ 
#ifndef IPOD
	         if (!execl ("/usr/bin/timidity", "/usr/bin/timidity", "-s", "22050", "-Od8M", "-idqqqq", filename, (char *)NULL))
#else
	         if (!execl ("/usr/bin/timidity", "/usr/bin/timidity", "-s", "22050", "-Od1M", "-idqqqq", filename, (char *)NULL))
#endif
		 {
      		 	pz_perror("Execl ");
			return TTK_MENU_DONOTHING;
		 }
	         _exit (1);
		break;
	 case -1: 
		 pz_perror ("Process didn't spawn right (vfork)"); 
		 break; 
	 default: 
                /* It looks like it launched. How frightening */

		ret = pz_new_window("MIDI Player", PZ_WINDOW_NORMAL);
    		wid = pz_add_widget(ret, draw_midihelper, event_midihelper);
		ttk_widget_set_timer(wid, 500);
		wid->timer = midihelper_loop;
		return pz_finish_window(ret);
	}


/* Podzilla is bending the will of space and time if it gets
** down this far.
** What more is podzilla capable of?
*/

	return TTK_MENU_DONOTHING;
}

void cleanup_midihelper() 
{

/* Not much to do here, since well.. I don't really do that much at all.
** Note the function that removes the association of midi's from the 
** file browser. Be polite and remember to disassociate yours if your
** helper module gets removed.
*/
   pz_browser_remove_handler(is_midi);
}

void init_midihelper() 
{

/* Put all of your start-up stuff here. I can't imagine what that may be
** except in my case making file-browser associations.
** The second parameter determines what to do when the module
** closes. 
*/
    midi_module = pz_register_module ("midihelper", cleanup_midihelper);

/* This function is the function that tells the file browser
** what to do you with your module.
** First parameter is your test to determine what kind of file 
** we're dealing with. The argument appears to be the full path of the thing
** you are clicking on in file browser (I believe). The argument has to
** return TRUE for it to continue onwards with the file.
** Second parameter is what you do with the file once its determined
** that your module is the one to be used. This should be a Twindow.
*/

    pz_browser_set_handler(is_midi, open_midi);
}

PZ_MOD_INIT (init_midihelper)

